import os
import pytest
import httpx
from unittest.mock import AsyncMock, MagicMock, patch
from dot_search.embed import make_openai_compatible_embed_fn, get_embedder


def make_mock_response(texts: list[str]) -> dict:
    return {
        "data": [
            {"index": i, "embedding": [float(i), 0.0, 0.0]}
            for i in range(len(texts))
        ]
    }


@pytest.fixture
def embedder():
    return make_openai_compatible_embed_fn(
        api_key="sk-test",
        model="text-embedding-3-small",
        base_url="https://api.openai.com/v1",
        dimension=1536,
    )


async def test_returns_one_vector_per_text(embedder):
    texts = ["hello", "world", "foo"]
    mock_response = AsyncMock()
    mock_response.json = lambda: make_mock_response(texts)
    mock_response.raise_for_status = lambda: None

    with patch("httpx.AsyncClient.post", return_value=mock_response):
        result = await embedder.fn(texts)

    assert len(result) == 3
    assert all(isinstance(v, list) for v in result)


async def test_vectors_are_ordered_by_index(embedder):
    texts = ["a", "b", "c"]
    # Return embeddings in reverse index order
    mock_response = AsyncMock()
    mock_response.json = lambda: {
        "data": [
            {"index": 2, "embedding": [2.0]},
            {"index": 0, "embedding": [0.0]},
            {"index": 1, "embedding": [1.0]},
        ]
    }
    mock_response.raise_for_status = lambda: None

    with patch("httpx.AsyncClient.post", return_value=mock_response):
        result = await embedder.fn(texts)

    assert result[0] == [0.0]
    assert result[1] == [1.0]
    assert result[2] == [2.0]


async def test_raises_on_http_error(embedder):
    mock_response = AsyncMock()
    mock_response.raise_for_status = lambda: (_ for _ in ()).throw(
        httpx.HTTPStatusError("500", request=None, response=None)
    )

    with patch("httpx.AsyncClient.post", return_value=mock_response):
        with pytest.raises(httpx.HTTPStatusError):
            await embedder.fn(["hello"])


async def test_sends_correct_payload(embedder):
    texts = ["test text"]
    mock_response = AsyncMock()
    mock_response.json = lambda: make_mock_response(texts)
    mock_response.raise_for_status = lambda: None

    with patch("httpx.AsyncClient.post", return_value=mock_response) as mock_post:
        await embedder.fn(texts)

    call_kwargs = mock_post.call_args
    payload = call_kwargs.kwargs["json"]
    assert payload["input"] == texts
    assert payload["model"] == "text-embedding-3-small"


def test_get_embedder_with_default_provider():
    """Test that provider='default' reads from environment variables."""
    os.environ["DOT__EMBED_API_KEY"] = "sk-test"
    os.environ["DOT__EMBED_BASE_URL"] = "https://api.openai.com/v1"
    os.environ["DOT__EMBED_MODEL"] = "text-embedding-3-small"
    os.environ["DOT__EMBED_DIMENSION"] = "1536"

    embedder = get_embedder("default", "ignored-model")

    assert embedder.dimension == 1536
    assert embedder.fn is not None


def test_get_embedder_with_openai_provider():
    """Test that provider='openai' creates an OpenAI-compatible embedder."""
    os.environ["DOT__EMBED_API_KEY"] = "sk-test"
    os.environ["DOT__EMBED_BASE_URL"] = "https://api.openai.com/v1"
    os.environ["DOT__EMBED_DIMENSION"] = "1536"

    embedder = get_embedder("openai", "text-embedding-3-large")

    assert embedder.dimension == 1536
    assert embedder.fn is not None


def test_get_embedder_with_unknown_provider():
    """Test that unknown providers raise ValueError."""
    with pytest.raises(ValueError, match="Unknown provider"):
        get_embedder("unknown", "model")


def test_get_embedder_openai_missing_env_vars():
    """Test that missing env vars for openai provider raise KeyError."""
    # Clear environment
    os.environ.pop("DOT__EMBED_API_KEY", None)
    os.environ.pop("DOT__EMBED_BASE_URL", None)
    os.environ.pop("DOT__EMBED_DIMENSION", None)

    with pytest.raises(KeyError):
        get_embedder("openai", "text-embedding-3-large")



import pytest


# Fake embed_fn — returns deterministic vectors based on text content.
# Used in all tests that need an embed_fn without hitting an API.
async def fake_embed_fn(texts: list[str]) -> list[list[float]]:
    # Returns a 4-dim vector where each component is based on character ordinals.
    # Same text always produces the same vector.
    def to_vec(text: str) -> list[float]:
        base = [float(ord(c)) for c in text[:4].ljust(4)]
        norm = sum(x**2 for x in base) ** 0.5 or 1.0
        return [x / norm for x in base]

    return [to_vec(t) for t in texts]


DB_URL = "postgresql+asyncpg://postgres:postgres@localhost/dot_search_test"



from __future__ import annotations

import os
from dataclasses import dataclass
from typing import Awaitable, Callable

import httpx

EmbedFn = Callable[[list[str]], Awaitable[list[list[float]]]]


@dataclass
class Embedder:
    fn: EmbedFn
    dimension: int


def make_openai_compatible_embed_fn(api_key: str, model: str, base_url: str, dimension: int) -> Embedder:
    url = base_url.rstrip("/") + "/embeddings"

    async def embed(texts: list[str]) -> list[list[float]]:
        async with httpx.AsyncClient() as client:
            response = await client.post(
                url,
                headers={"Authorization": f"Bearer {api_key}"},
                json={"input": texts, "model": model},
            )
            response.raise_for_status()
            data = response.json()
            items = sorted(data["data"], key=lambda x: x["index"])
            return [item["embedding"] for item in items]

    return Embedder(fn=embed, dimension=dimension)


def get_embedder(provider: str, model: str) -> Embedder:
    if provider in ("default", "openai"):
        return make_openai_compatible_embed_fn(
            api_key=os.environ["DOT__EMBED_API_KEY"],
            model=os.environ["DOT__EMBED_MODEL"] if model in ("default", "") else model,
            base_url=os.environ["DOT__EMBED_BASE_URL"],
            dimension=int(os.environ["DOT__EMBED_DIMENSION"]),
        )

    raise ValueError(f"Unknown provider: {provider!r}. Supported: 'default', 'openai'")



from typing import Any

from dot_search.types import SearchResult


def reciprocal_rank_fusion(
    ranked_lists: list[list[SearchResult]],
    weights: list[float],
    k: int = 60,
) -> list[SearchResult]:
    scores: dict[str, float] = {}
    ids: dict[str, Any] = {}
    rows: dict[str, dict] = {}

    for ranked, weight in zip(ranked_lists, weights):
        for rank, result in enumerate(ranked):
            key = str(result.id)
            scores[key] = scores.get(key, 0.0) + weight / (k + rank + 1)
            ids[key] = result.id
            rows[key] = result.row

    sorted_keys = sorted(scores, key=lambda x: scores[x], reverse=True)
    return [
        SearchResult(id=ids[key], score=scores[key], row=rows[key])
        for key in sorted_keys
    ]



from __future__ import annotations

import dataclasses
import json

from dot_search.embed import get_embedder
from dot_search.fusion import reciprocal_rank_fusion
from dot_search.store import make_store
from dot_search.store.protocol import BM25Search, ExactSearch, SubSearch
from dot_search.types import SearchConfig, SearchResult, TableConfig


class SearchEngine:
    def __init__(self, db_url: str) -> None:
        self._store = make_store(db_url)

    async def _ensure_config_table(self) -> None:
        await self._store.ensure_config_table()

    async def _save_config(self, config: TableConfig) -> None:
        data = dataclasses.asdict(config)
        await self._store.save_config(config.index_id, json.dumps(data))

    async def _load_config(self, index_id: str) -> TableConfig:
        raw = await self._store.load_config(index_id)
        if raw is None:
            raise ValueError(f"No config found for index_id '{index_id}'. Call index(config=...) first.")
        return TableConfig.from_json(raw)

    async def index(self, config: TableConfig) -> None:
        await self._ensure_config_table()
        await self._save_config(config)
        await self._run_indexing(config)

    async def _run_indexing(self, config: TableConfig) -> None:
        existing_columns = set(await self._store.get_columns(config.table))

        for embedding_config in config.embeddings:
            if embedding_config.source_column not in existing_columns:
                raise ValueError(
                    f"EmbeddingConfig source_column {embedding_config.source_column!r} "
                    f"does not exist in table {config.table!r}. "
                    f"Available columns: {sorted(existing_columns)}"
                )

        for bm25cfg in config.bm25:
            if bm25cfg.source_column not in existing_columns:
                raise ValueError(
                    f"BM25Config source_column {bm25cfg.source_column!r} "
                    f"does not exist in table {config.table!r}. "
                    f"Available columns: {sorted(existing_columns)}"
                )

        for embedding_config in config.embeddings:
            embedder = get_embedder(embedding_config.provider, embedding_config.model)
            target_col = embedding_config.target_column
            rows = await self._store.read_column(config.table, config.pk_column, embedding_config.source_column)

            if not rows:
                continue

            ids = [r[0] for r in rows]
            texts = [r[1] for r in rows]

            await self._store.add_embedding_column(config.table, target_col, embedder.dimension)

            checked = False
            for batch_start in range(0, len(texts), config.batch_size):
                batch_texts = texts[batch_start : batch_start + config.batch_size]
                batch_ids = ids[batch_start : batch_start + config.batch_size]
                batch_vecs = await embedder.fn(batch_texts)
                if not checked and batch_vecs and len(batch_vecs[0]) != embedder.dimension:
                    raise ValueError(
                        f"Embedder dimension mismatch: configured {embedder.dimension}, "
                        f"but fn returned vectors of size {len(batch_vecs[0])}."
                    )
                checked = True
                await self._store.write_embeddings(
                    config.table,
                    target_col,
                    config.pk_column,
                    list(zip(batch_ids, batch_vecs)),
                )

        if config.bm25:
            await self._store.add_bm25_index(config.table, [t.source_column for t in config.bm25])

    async def search(
        self,
        query: str,
        index_id: str,
        config: SearchConfig | None = None,
    ) -> list[SearchResult]:
        if config is None:
            config = SearchConfig()

        await self._ensure_config_table()
        table_config = await self._load_config(index_id)

        strategy = config.strategy
        limit = config.limit
        weights_map = config.weights

        subsearches: list[SubSearch] = []
        weights: list[float] = []

        if strategy in ("vector", "hybrid"):
            for embedding_config in table_config.embeddings:
                embedder = get_embedder(embedding_config.provider, embedding_config.model)
                target_col = embedding_config.target_column
                w = weights_map.get(embedding_config.key, 1.0)
                query_embedding = (await embedder.fn([query]))[0]
                subsearches.append(SubSearch(
                    embedding_column=target_col,
                    query_embedding=query_embedding,
                    weight=w,
                ))
                weights.append(w)

        bm25searches = []
        if strategy in ("bm25", "hybrid"):
            if strategy == "bm25" and not table_config.bm25:
                raise ValueError(f"strategy='bm25' but index '{index_id}' has no bm25 configured.")
            bm25searches = [
                BM25Search(
                    column=t.source_column,
                    query=query,
                    weight=weights_map.get(t.key, 1.0),
                )
                for t in table_config.bm25
            ]
            weights.extend(s.weight for s in bm25searches)

        exact_searches = []
        if strategy in ("exact", "hybrid"):
            if strategy == "exact" and not table_config.exact:
                raise ValueError(f"strategy='exact' but index '{index_id}' has no exact config.")
            exact_searches = [
                ExactSearch(
                    column=t.source_column,
                    query=query,
                    weight=weights_map.get(t.key, 1.0),
                )
                for t in table_config.exact
            ]
            weights.extend(s.weight for s in exact_searches)

        ranked_lists = await self._store.search(
            table_config.table,
            table_config.pk_column,
            subsearches,
            bm25searches,
            exact_searches,
            config.where,
            limit,
        )

        return reciprocal_rank_fusion(ranked_lists, weights)[:limit]

    async def close(self) -> None:
        await self._store.close()



import asyncio
import os

import sqlalchemy as sa
from sqlalchemy.ext.asyncio import create_async_engine

from dot_search import (
    BM25Config,
    EmbeddingConfig,
    TableConfig,
    SearchEngine,
    serialize_row,
)

DB_URL = os.environ.get("DOT__DB_URL", "postgresql+asyncpg://postgres:postgres@localhost:5433/postgres")

PRODUCTS = [
    (1,  "Wireless Noise-Cancelling Headphones", "Premium over-ear headphones with active noise cancellation, 30-hour battery, and foldable design for travel.", "electronics", 249.99, True,  "audio bluetooth travel"),
    (2,  "Mechanical Keyboard",                  "Tenkeyless mechanical keyboard with Cherry MX Blue switches, RGB backlighting, and aluminum frame.",               "electronics", 129.00, True,  "typing gaming peripherals"),
    (3,  "Cast Iron Skillet",                    "Pre-seasoned 12-inch cast iron skillet perfect for searing steaks, baking cornbread, and stovetop cooking.",       "kitchen",      39.95, True,  "cooking baking stovetop"),
    (4,  "Yoga Mat",                             "Eco-friendly 6mm thick yoga mat made from natural rubber, non-slip surface, ideal for hot yoga and pilates.",      "fitness",      55.00, True,  "exercise wellness stretching"),
    (5,  "Espresso Machine",                     "Compact semi-automatic espresso machine with 15-bar pump, steam wand for milk frothing, and stainless steel body.", "kitchen",    199.00, True,  "coffee espresso brewing"),
    (6,  "Running Shoes",                        "Lightweight neutral running shoes with responsive foam midsole, breathable mesh upper, and reflective details.",   "footwear",     89.99, False, "running athletics outdoor"),
    (7,  "Portable Solar Charger",               "20000 mAh solar power bank with dual USB-A and USB-C outputs, waterproof casing, and foldable solar panels.",     "electronics",  44.99, True,  "solar outdoor camping"),
    (8,  "Stainless Steel Water Bottle",         "Vacuum-insulated 32oz water bottle keeps drinks cold 24 hours or hot 12 hours, leak-proof lid, BPA-free.",        "outdoors",     27.50, True,  "hydration hiking camping"),
    (9,  "Bamboo Cutting Board",                 "Large end-grain bamboo cutting board with juice groove, non-slip feet, and antimicrobial surface.",               "kitchen",      34.00, True,  "cooking food prep kitchen"),
    (10, "4K Action Camera",                     "Waterproof action camera recording 4K at 60fps with image stabilization, wide-angle lens, and Wi-Fi streaming.",  "electronics",  79.99, True,  "video photography adventure"),
    (11, "Foam Roller",                          "High-density EVA foam roller for deep tissue massage, muscle recovery, and myofascial release post-workout.",     "fitness",      22.00, True,  "recovery massage sports"),
    (12, "Electric Kettle",                      "Variable temperature 1.7L electric kettle with gooseneck spout, stainless steel interior, ideal for pour-over coffee and tea.", "kitchen", 49.00, False, "tea coffee kitchen"),
    (13, "Smart Watch",                          "GPS smartwatch with heart rate monitor, sleep tracking, 5-day battery, and smartphone notifications.",            "electronics", 199.95, True,  "fitness health wearable"),
    (14, "Backpacking Tent",                     "Ultralight 2-person backpacking tent weighing under 2kg, freestanding, with waterproof rainfly and vestibule.",   "outdoors",    159.00, True,  "camping hiking shelter"),
    (15, "Resistance Bands Set",                 "Set of 5 latex resistance bands with varying tension levels for strength training and physical therapy.",         "fitness",      18.99, True,  "exercise rehab strength"),
    (16, "Ceramic Pour-Over Coffee Dripper",     "Handcrafted ceramic dripper compatible with Chemex and standard carafes, produces clean bright coffee flavor.",   "kitchen",      28.00, True,  "coffee brewing pour-over"),
    (17, "USB-C Hub",                            "7-in-1 USB-C hub with HDMI 4K, 3x USB-A, SD card reader, USB-C PD charging, and Ethernet port.",               "electronics",  35.00, True,  "laptop accessories connectivity"),
    (18, "Hiking Boots",                         "Waterproof mid-cut hiking boots with Vibram outsole, ankle support, and Gore-Tex lining for wet terrain.",       "footwear",    149.00, False, "hiking outdoor waterproof"),
    (19, "Dutch Oven",                           "5.5-quart enameled cast iron Dutch oven for slow braising, soups, and artisan bread baking.",                    "kitchen",      79.00, True,  "cooking baking kitchen"),
    (20, "Noise Machine",                        "White noise machine with 30 sounds, timer, and night light, for better sleep and focus in open offices.",        "electronics",  39.99, True,  "sleep focus relaxation"),
    (21, "Climbing Harness",                     "Lightweight sport climbing harness with padded waist and leg loops, adjustable buckles, and gear loops.",        "outdoors",     65.00, True,  "climbing safety rock"),
    (22, "Cold Brew Coffee Maker",               "1-gallon glass cold brew system with fine mesh filter, produces smooth low-acid coffee concentrate overnight.",  "kitchen",      24.99, True,  "coffee cold-brew beverage"),
    (23, "Laptop Stand",                         "Adjustable aluminum laptop stand with 6 height settings, foldable design, and compatibility with 10-17 inch laptops.", "electronics", 29.99, True, "ergonomic desk accessories"),
    (24, "Trekking Poles",                       "Collapsible carbon fiber trekking poles with ergonomic cork grips, flick locks, and interchangeable tips.",      "outdoors",     75.00, False, "hiking trekking poles"),
    (25, "Air Purifier",                         "True HEPA air purifier covering up to 500 sq ft, removes 99.97% of allergens, dust, pollen, and pet dander.",   "home",        119.00, True,  "health air quality home"),
]

ARTICLES = [
    (1,  "The Rise of Quantum Computing",           "Quantum computers leverage qubits in superposition to solve optimization problems exponentially faster than classical machines. IBM and Google have demonstrated quantum advantage in specific tasks.",   "Alice Chen",        2023, "technology"),
    (2,  "Mediterranean Diet and Longevity",         "Studies spanning decades show that populations adhering to the Mediterranean diet, rich in olive oil, legumes, and fish, have significantly lower rates of cardiovascular disease and longer lifespans.", "Dr. Marco Rossi",   2022, "health"),
    (3,  "Reforestation Efforts in the Amazon",     "Brazil's federal reforestation program aims to restore 12 million hectares of degraded rainforest by 2030, sequestering carbon and protecting endemic species in the world's largest tropical ecosystem.", "Ana Lima",          2023, "environment"),
    (4,  "Machine Learning in Drug Discovery",      "AI models trained on molecular structures can predict protein-ligand binding affinities with remarkable accuracy, cutting years off early-stage pharmaceutical research and reducing development costs.",  "Dr. Sarah Kim",     2024, "science"),
    (5,  "The Art of Sourdough Fermentation",       "Wild-fermented sourdough bread relies on a complex microbiome of lactobacilli and wild yeasts. Long cold fermentation develops flavor complexity and improves digestibility compared to commercial yeast.", "Jean-Paul Blanc",   2021, "food"),
    (6,  "Deep Sea Exploration Breakthroughs",      "New pressure-resistant ROVs are mapping previously unexplored hadal zones at depths exceeding 10000 meters, discovering chemosynthetic ecosystems and novel extremophile organisms.",                      "Dr. Mia Torres",    2023, "science"),
    (7,  "Urban Vertical Farming",                  "Vertical farms using LED lighting and hydroponic systems can grow leafy greens year-round with 95% less water than traditional agriculture, addressing food security in dense urban environments.",         "James Wu",          2022, "food"),
    (8,  "The History of the Silk Road",            "The ancient Silk Road connected Han dynasty China to Rome through 4000 miles of overland and maritime trade routes, facilitating exchange of silk, spice, glass, and ideas across Eurasia for centuries.", "Prof. Omar Said",   2020, "history"),
    (9,  "Training for an Ultramarathon",           "Ultramarathon preparation requires base-building phases of 60-80 miles per week, back-to-back long runs, and targeted nutrition strategies including fat adaptation and carbohydrate periodization.",       "Lisa Nakamura",     2023, "fitness"),
    (10, "Fermentation and Gut Health",             "Fermented foods like kimchi, kefir, and sauerkraut introduce beneficial bacteria to the gut microbiome, reducing systemic inflammation and improving mental health outcomes through the gut-brain axis.",   "Dr. Priya Shah",    2022, "health"),
    (11, "Exploring Patagonia's Torres del Paine",  "Patagonia's W Trek winds through granite towers, teal glacial lakes, and windswept pampas. The five-day route crosses hanging glaciers and offers encounters with guanacos and Andean condors.",            "Carlos Mendez",     2023, "travel"),
    (12, "The Neuroscience of Sleep",               "During NREM slow-wave sleep the hippocampus replays memories for consolidation to the neocortex. Chronic sleep deprivation impairs synaptic plasticity and accelerates amyloid-beta accumulation linked to Alzheimer's.", "Dr. Emily Foster", 2024, "science"),
    (13, "Ramen: From Street Food to Global Icon",  "Ramen originated as Chinese-style wheat noodles adapted in postwar Japan. Regional varieties like Sapporo miso, Hakata tonkotsu, and Tokyo shoyu reflect local ingredients and evolving culinary traditions.", "Kenji Watanabe",   2021, "food"),
    (14, "Renewable Energy Storage Challenges",     "Intermittency of solar and wind power demands scalable storage. Grid-scale lithium-ion batteries, pumped hydro, and emerging iron-air batteries each present different cost-capacity-lifespan tradeoffs.",  "Dr. Fatima Al-Hassan", 2023, "technology"),
    (15, "Kyoto's Hidden Temples",                  "Beyond Kinkakuji and Fushimi Inari lie lesser-known gems: Daitoku-ji's moss gardens, Fushimi Momoyama's samurai heritage, and Kurama's mountain shrine accessible only by narrow forest trail.",           "Yuki Tanaka",       2022, "travel"),
]


async def setup_tables(engine: sa.ext.asyncio.AsyncEngine) -> None:
    async with engine.begin() as conn:
        await conn.execute(sa.text("DROP TABLE IF EXISTS products CASCADE"))
        await conn.execute(sa.text("DROP TABLE IF EXISTS articles CASCADE"))
        await conn.execute(sa.text("DROP TABLE IF EXISTS ds_configs CASCADE"))
        await conn.execute(sa.text("""
            CREATE TABLE products (
                id SERIAL PRIMARY KEY,
                name TEXT NOT NULL,
                description TEXT NOT NULL,
                category TEXT NOT NULL,
                price FLOAT NOT NULL,
                in_stock BOOLEAN NOT NULL,
                tags TEXT NOT NULL
            )
        """))
        await conn.execute(sa.text("""
            CREATE TABLE articles (
                id SERIAL PRIMARY KEY,
                title TEXT NOT NULL,
                body TEXT NOT NULL,
                author TEXT NOT NULL,
                published_year INT NOT NULL,
                topic TEXT NOT NULL
            )
        """))

    async with engine.begin() as conn:
        await conn.execute(
            sa.text("INSERT INTO products (id, name, description, category, price, in_stock, tags) VALUES (:id, :name, :description, :category, :price, :in_stock, :tags)"),
            [{"id": r[0], "name": r[1], "description": r[2], "category": r[3], "price": r[4], "in_stock": r[5], "tags": r[6]} for r in PRODUCTS],
        )
        await conn.execute(
            sa.text("INSERT INTO articles (id, title, body, author, published_year, topic) VALUES (:id, :title, :body, :author, :published_year, :topic)"),
            [{"id": r[0], "title": r[1], "body": r[2], "author": r[3], "published_year": r[4], "topic": r[5]} for r in ARTICLES],
        )

    print(f"  Inserted {len(PRODUCTS)} products and {len(ARTICLES)} articles.")


def print_results(results, text_key: str) -> None:
    for i, r in enumerate(results, 1):
        text_preview = r.row.get(text_key, "")[:80].replace("\n", " ")
        print(f"  {i}. [id={r.id}] score={r.score:.4f}  {text_preview}")


async def main() -> None:
    print("=" * 70)
    print("dot-search integration test")
    print("=" * 70)

    # ------------------------------------------------------------------ setup
    print("\n--- Setting up tables ---")
    raw_engine = create_async_engine(DB_URL)
    await setup_tables(raw_engine)
    await raw_engine.dispose()

    engine = SearchEngine(DB_URL)

    # --------------------------------------------------------------- indexing
    print("\n--- Indexing products (ds_desc + ds_name, BM25 on description) ---")
    products_config = TableConfig(
        table="products",
        embeddings=[
            EmbeddingConfig(target_column="ds_products_description_default", source_column="description"),
            EmbeddingConfig(target_column="ds_products_name_default", source_column="name"),
        ],
        pk_column="id",
        bm25=[BM25Config(source_column="description")],
        batch_size=50,
    )
    await engine.index(config=products_config)
    print("  products indexed.")

    print("\n--- Indexing articles (ds_body, BM25 on body) ---")
    articles_config = TableConfig(
        table="articles",
        embeddings=[
            EmbeddingConfig(target_column="ds_articles_body_default", source_column="body"),
        ],
        pk_column="id",
        bm25=[BM25Config(source_column="body")],
        batch_size=50,
    )
    await engine.index(config=articles_config)
    print("  articles indexed.")

    # --------------------------------------------------------------- searches
    print("\n" + "=" * 70)
    print("SEARCH QUERIES")
    print("=" * 70)

    # Q1: products — audio / headphones
    print("\n[Q1] Products: 'over-ear headphones with noise cancellation'")
    results = await engine.search(
        "over-ear headphones with noise cancellation",
        "products",
        limit=5,
    )
    print_results(results, "description")
    assert results, "Expected at least one result"
    assert results[0].id == 1, f"Top result should be headphones (id=1), got id={results[0].id}"
    assert [r.score for r in results] == sorted([r.score for r in results], reverse=True)

    # Q2: products — coffee brewing
    print("\n[Q2] Products: 'brewing espresso coffee at home'")
    results = await engine.search(
        "brewing espresso coffee at home",
        "products",
        limit=5,
    )
    print_results(results, "description")
    assert results
    assert [r.score for r in results] == sorted([r.score for r in results], reverse=True)
    top_ids = [r.id for r in results[:3]]
    coffee_ids = {5, 12, 16, 22}
    assert coffee_ids & set(top_ids), f"Expected a coffee product in top 3, got {top_ids}"

    # Q3: products with WHERE in_stock = true
    print("\n[Q3] Products (in_stock=true): 'outdoor hiking gear'")
    results = await engine.search(
        "outdoor hiking gear",
        "products",
        limit=5,
        where="in_stock = true",
    )
    print_results(results, "description")
    assert results
    for r in results:
        assert r.row["in_stock"] is True, f"Result id={r.id} has in_stock={r.row['in_stock']}, expected True"
    assert [r.score for r in results] == sorted([r.score for r in results], reverse=True)

    # Q4: products with WHERE price < 50
    print("\n[Q4] Products (price<50): 'kitchen cooking tools'")
    results = await engine.search(
        "kitchen cooking tools",
        "products",
        limit=5,
        where="price < 50",
    )
    print_results(results, "description")
    assert results
    for r in results:
        assert r.row["price"] < 50, f"Result id={r.id} has price={r.row['price']}, expected < 50"
    assert [r.score for r in results] == sorted([r.score for r in results], reverse=True)

    # Q5: products with WHERE category = 'electronics'
    print("\n[Q5] Products (category=electronics): 'wireless connectivity device'")
    results = await engine.search(
        "wireless connectivity device",
        "products",
        limit=5,
        where="category = 'electronics'",
    )
    print_results(results, "description")
    assert results
    for r in results:
        assert r.row["category"] == "electronics", f"Result id={r.id} has category={r.row['category']!r}"
    assert [r.score for r in results] == sorted([r.score for r in results], reverse=True)

    # Q6: articles — quantum computing
    print("\n[Q6] Articles: 'quantum computers and computational advantage'")
    results = await engine.search(
        "quantum computers and computational advantage",
        "articles",
        limit=5,
    )
    print_results(results, "body")
    assert results
    assert results[0].id == 1, f"Top result should be quantum computing article (id=1), got id={results[0].id}"
    assert [r.score for r in results] == sorted([r.score for r in results], reverse=True)

    # Q7: articles — gut health and fermented food
    print("\n[Q7] Articles: 'fermented foods gut microbiome health'")
    results = await engine.search(
        "fermented foods gut microbiome health",
        "articles",
        limit=5,
    )
    print_results(results, "body")
    assert results
    top_ids = [r.id for r in results[:2]]
    assert 10 in top_ids or 5 in top_ids, f"Expected fermentation article (id=5 or 10) in top 2, got {top_ids}"
    assert [r.score for r in results] == sorted([r.score for r in results], reverse=True)

    # Q8: articles — travel in Japan
    print("\n[Q8] Articles: 'travel Japan temples ancient culture'")
    results = await engine.search(
        "travel Japan temples ancient culture",
        "articles",
        limit=5,
    )
    print_results(results, "body")
    assert results
    assert results[0].id == 15, f"Top result should be Kyoto temples article (id=15), got id={results[0].id}"
    assert [r.score for r in results] == sorted([r.score for r in results], reverse=True)

    # Q9: BM25 search on articles
    print("\n[Q9] Articles BM25: 'fermentation'")
    results = await engine.search(
        "fermentation",
        "articles",
        strategy="bm25",
        limit=5,
    )
    print_results(results, "body")
    assert results, "BM25 search returned no results"
    top_ids = [r.id for r in results[:2]]
    assert 5 in top_ids or 10 in top_ids, f"Expected fermentation article in top 2, got {top_ids}"
    assert [r.score for r in results] == sorted([r.score for r in results], reverse=True)

    # Q10: hybrid search — vector + BM25 fused via RRF
    print("\n[Q10] Articles hybrid (vector + BM25): 'sourdough bread microbiome'")
    results = await engine.search(
        "sourdough bread microbiome",
        "articles",
        limit=5,
    )
    print_results(results, "body")
    assert results
    assert results[0].id == 5, f"Top hybrid result should be sourdough article (id=5), got id={results[0].id}"
    assert [r.score for r in results] == sorted([r.score for r in results], reverse=True)

    # Q11: source_column=None — auto-serialize all columns
    print("\n[Q11] Products auto-serialize (source_column=None): 'lightweight outdoor shelter'")
    products_all_config = TableConfig(
        table="products",
        index_id="products_all",
        embeddings=[
            EmbeddingConfig(target_column="ds_products_all_auto_default", source_column=None),
        ],
        pk_column="id",
    )
    await engine.index(config=products_all_config)
    results = await engine.search(
        "lightweight outdoor shelter",
        "products_all",
        limit=3,
    )
    print_results(results, "description")
    assert results
    assert results[0].id == 14, f"Top result should be backpacking tent (id=14), got id={results[0].id}"

    # Q12: small batch_size forces multiple batches — re-index articles with batch_size=4
    print("\n[Q12] Re-indexing articles with batch_size=4 (forces multiple batches over 15 rows)")
    articles_rebatch_config = TableConfig(
        table="articles",
        embeddings=[
            EmbeddingConfig(target_column="ds_articles_body_default", source_column="body"),
        ],
        pk_column="id",
        bm25=[BM25Config(source_column="body")],
        batch_size=4,
    )
    await engine.index(config=articles_rebatch_config)
    results = await engine.search(
        "quantum computers and computational advantage",
        "articles",
        limit=3,
    )
    print_results(results, "body")
    assert results[0].id == 1, f"After re-index, top result should still be id=1, got {results[0].id}"

    # Q13: BM25 + WHERE filter
    print("\n[Q13] Articles BM25 + WHERE (topic=food): 'fermentation'")
    results = await engine.search(
        "fermentation",
        "articles",
        strategy="bm25",
        limit=5,
        where="topic = 'food'",
    )
    print_results(results, "body")
    assert results
    for r in results:
        assert r.row["topic"] == "food", f"Result id={r.id} has topic={r.row['topic']!r}, expected 'food'"
    assert results[0].id == 5, f"Top BM25+WHERE result should be sourdough article (id=5), got {results[0].id}"

    # Q14: divergent RRF — name-biased vs description-biased targets
    print("\n[Q14] RRF fusion — name-biased vs description-biased targets")
    await engine.index(config=TableConfig(
        table="products",
        index_id="products_name_only",
        embeddings=[EmbeddingConfig(target_column="ds_products_name_only_name_default", source_column="name")],
        pk_column="id",
    ))
    name_only = await engine.search("Dutch Oven", "products_name_only", limit=5)

    await engine.index(config=TableConfig(
        table="products",
        index_id="products_desc_only",
        embeddings=[EmbeddingConfig(target_column="ds_products_desc_only_description_default", source_column="description")],
        pk_column="id",
    ))
    desc_only = await engine.search("Dutch Oven", "products_desc_only", limit=5)

    await engine.index(config=TableConfig(
        table="products",
        index_id="products_fused",
        embeddings=[
            EmbeddingConfig(target_column="ds_products_fused_name_default", source_column="name"),
            EmbeddingConfig(target_column="ds_products_fused_description_default", source_column="description"),
        ],
        pk_column="id",
    ))
    fused = await engine.search("Dutch Oven", "products_fused", limit=5)

    print(f"  name-only top3: {[r.id for r in name_only[:3]]}")
    print(f"  desc-only top3: {[r.id for r in desc_only[:3]]}")
    print(f"  fused     top3: {[r.id for r in fused[:3]]}")
    assert name_only[0].id == 19, f"name-only top should be Dutch Oven (id=19), got {name_only[0].id}"
    assert fused[0].id == 19, f"fused top should be Dutch Oven (id=19), got {fused[0].id}"
    fused_top_score = fused[0].score
    name_top_score = name_only[0].score
    assert fused_top_score != name_top_score, "Fused score should differ from single-target score"

    # ----------------------------------------------------------------- cleanup
    await engine.close()

    print("\n" + "=" * 70)
    print("All assertions passed.")
    print("=" * 70)


if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations

import sqlite3
import struct
from typing import Any

import aiosqlite

from dot_search.store.protocol import BM25Search, ExactSearch, SubSearch, validate_embedding_column
from dot_search.types import SearchResult


class SQLiteStore:
    def __init__(self, db_url: str) -> None:
        import sqlite_vec
        path = db_url.replace("sqlite+aiosqlite:///", "").replace("sqlite+aiosqlite://", "")
        if not path or path == ":memory:":
            path = ":memory:"
        self._path = path
        self._vec_path = sqlite_vec.loadable_path()
        self._conn: aiosqlite.Connection | None = None

    async def _get_conn(self) -> aiosqlite.Connection:
        if self._conn is None:
            self._conn = await aiosqlite.connect(self._path)
            self._conn.row_factory = sqlite3.Row
            await self._conn.enable_load_extension(True)
            await self._conn.load_extension(self._vec_path)
            await self._conn.enable_load_extension(False)
        return self._conn

    async def add_embedding_column(self, table: str, column: str, dimension: int) -> None:
        validate_embedding_column(column)
        conn = await self._get_conn()
        await conn.execute(f"ALTER TABLE {table} ADD COLUMN {column} BLOB")
        await conn.commit()

    async def add_text_column(self, table: str, column: str) -> None:
        conn = await self._get_conn()
        await conn.execute(f"ALTER TABLE {table} ADD COLUMN {column} TEXT")
        await conn.commit()

    async def write_texts(self, table: str, column: str, pk_column: str, rows: list[tuple[Any, str]]) -> None:
        if not rows:
            return
        conn = await self._get_conn()
        for id_, text in rows:
            await conn.execute(
                f"UPDATE {table} SET {column} = ? WHERE {pk_column} = ?",
                (text, id_),
            )
        await conn.commit()

    async def add_bm25_index(self, table: str, columns: list[str]) -> None:
        pass

    async def write_embeddings(
        self,
        table: str,
        column: str,
        pk_column: str,
        rows: list[tuple[Any, list[float]]],
    ) -> None:
        if not rows:
            return
        conn = await self._get_conn()
        for id_, vec in rows:
            blob = struct.pack(f"{len(vec)}f", *vec)
            await conn.execute(
                f"UPDATE {table} SET {column} = ? WHERE {pk_column} = ?",
                (blob, id_),
            )
        await conn.commit()

    async def read_column(
        self,
        table: str,
        pk_column: str,
        source_column: str,
    ) -> list[tuple[Any, str]]:
        conn = await self._get_conn()
        cursor = await conn.execute(
            f"SELECT {pk_column}, {source_column} FROM {table}"
        )
        rows = await cursor.fetchall()
        return [(row[0], str(row[1]) if row[1] is not None else "") for row in rows]

    async def search(
        self,
        table: str,
        pk_column: str,
        subsearches: list[SubSearch],
        bm25searches: list[BM25Search],
        exact_searches: list[ExactSearch],
        where: str | None,
        limit: int,
    ) -> list[list[SearchResult]]:
        where_clause = f"AND {where}" if where else ""
        conn = await self._get_conn()
        results = []

        # Get payload columns (non-ds_)
        cursor = await conn.execute(f"SELECT * FROM {table} LIMIT 0")
        all_columns = [desc[0] for desc in cursor.description]
        payload_cols = [c for c in all_columns if not c.startswith("ds_")]

        for sub in subsearches:
            dim = len(sub.query_embedding)
            blob = struct.pack(f"{dim}f", *sub.query_embedding)

            select_cols = ", ".join(f"t.{c}" for c in payload_cols)
            sql = f"""
                SELECT {select_cols},
                       vec_distance_cosine(t.{sub.embedding_column}, ?) AS distance
                FROM {table} t
                WHERE t.{sub.embedding_column} IS NOT NULL
                {where_clause}
                ORDER BY distance ASC
                LIMIT ?
            """
            cursor = await conn.execute(sql, (blob, limit))
            rows = await cursor.fetchall()
            col_names = [desc[0] for desc in cursor.description]

            search_results = []
            for row in rows:
                row_dict = dict(zip(col_names, row))
                distance = row_dict.pop("distance")
                score = 1.0 - float(distance)
                id_val = row_dict[pk_column]
                search_results.append(SearchResult(id=id_val, score=score, row=row_dict))
            results.append(search_results)

        for exact in exact_searches:
            select_cols = ", ".join(payload_cols)
            sql = f"""
                SELECT {select_cols}
                FROM {table}
                WHERE {exact.column} LIKE ?
                {where_clause}
                LIMIT ?
            """
            cursor = await conn.execute(sql, (f"%{exact.query}%", limit))
            rows = await cursor.fetchall()
            col_names = [desc[0] for desc in cursor.description]

            search_results = []
            for row in rows:
                row_dict = dict(zip(col_names, row))
                id_val = row_dict[pk_column]
                search_results.append(SearchResult(id=id_val, score=1.0, row=row_dict))
            results.append(search_results)

        return results

    async def get_columns(self, table: str) -> list[str]:
        conn = await self._get_conn()
        cursor = await conn.execute(f"PRAGMA table_info({table})")
        rows = await cursor.fetchall()
        return [row[1] for row in rows]

    async def ensure_config_table(self) -> None:
        conn = await self._get_conn()
        await conn.execute(
            """
            CREATE TABLE IF NOT EXISTS ds_configs (
                table_name TEXT PRIMARY KEY,
                config_json TEXT NOT NULL
            )
            """
        )
        await conn.commit()

    async def save_config(self, index_id: str, config_json: str) -> None:
        conn = await self._get_conn()
        await conn.execute(
            "INSERT INTO ds_configs (table_name, config_json) VALUES (?, ?) ON CONFLICT (table_name) DO UPDATE SET config_json = excluded.config_json",
            (index_id, config_json),
        )
        await conn.commit()

    async def load_config(self, index_id: str) -> str | None:
        conn = await self._get_conn()
        cursor = await conn.execute(
            "SELECT config_json FROM ds_configs WHERE table_name = ?",
            (index_id,),
        )
        row = await cursor.fetchone()
        return row[0] if row else None

    async def close(self) -> None:
        if self._conn is not None:
            await self._conn.close()
            self._conn = None



"""
Validation tests:
1. EmbeddingConfig.target_column — ds_ prefix derived from key
2. validate_embedding_column — ds_ prefix in Store
3. Embedder dimension mismatch in engine._run_indexing
4. Column existence checks for EmbeddingConfig.source_column and BM25Config.source_column
5. TableConfig duplicate key validation
"""
from unittest.mock import patch

import pytest
from dot_search.types import BM25Config, EmbeddingConfig, ExactConfig, TableConfig
from dot_search.store.sqlite import SQLiteStore
from dot_search.embed import Embedder
from dot_search import SearchEngine
from conftest import fake_embed_fn


SQLITE_URL = "sqlite+aiosqlite://"
FAKE_EMBEDDER = Embedder(fn=fake_embed_fn, dimension=4)


# ---------------------------------------------------------------------------
# EmbeddingConfig target_column derivation (sync, no fixtures needed)
# ---------------------------------------------------------------------------

def test_embedding_config_target_column_from_key():
    cfg = EmbeddingConfig(key="body_emb", source_column="body")
    assert cfg.target_column == "ds_body_emb"


# ---------------------------------------------------------------------------
# TableConfig duplicate key validation
# ---------------------------------------------------------------------------

def test_table_config_rejects_duplicate_keys():
    with pytest.raises(ValueError, match="Duplicate keys"):
        TableConfig(
            table="docs",
            embeddings=[EmbeddingConfig(key="body", source_column="body")],
            exact=[ExactConfig(key="body", source_column="body")],
        )


def test_table_config_accepts_unique_keys():
    config = TableConfig(
        table="docs",
        embeddings=[EmbeddingConfig(key="body_emb", source_column="body")],
        bm25=[BM25Config(key="body_bm25", source_column="body")],
        exact=[ExactConfig(key="body_exact", source_column="body")],
    )
    assert config.table == "docs"


# ---------------------------------------------------------------------------
# SQLiteStore.add_embedding_column validation (async, needs a table)
# ---------------------------------------------------------------------------

@pytest.fixture
async def store():
    s = SQLiteStore(db_url=SQLITE_URL)
    yield s
    await s.close()


@pytest.fixture
async def table_with_data(store):
    conn = await store._get_conn()
    await conn.execute("CREATE TABLE IF NOT EXISTS docs (id TEXT PRIMARY KEY, body TEXT)")
    await conn.execute("INSERT OR IGNORE INTO docs (id, body) VALUES ('d1', 'hello world')")
    await conn.commit()
    yield
    await conn.execute("DROP TABLE IF EXISTS docs")
    await conn.commit()


async def test_add_embedding_column_rejects_column_without_ds_prefix(store, table_with_data):
    with pytest.raises(ValueError):
        await store.add_embedding_column("docs", "body_emb", dimension=4)


async def test_add_embedding_column_rejects_arbitrary_name(store, table_with_data):
    with pytest.raises(ValueError):
        await store.add_embedding_column("docs", "embedding", dimension=4)


async def test_add_embedding_column_accepts_valid_ds_prefix(store, table_with_data):
    await store.add_embedding_column("docs", "ds_body_emb", dimension=4)
    conn = await store._get_conn()
    cursor = await conn.execute("PRAGMA table_info(docs)")
    rows = await cursor.fetchall()
    columns = {row[1] for row in rows}
    assert "ds_body_emb" in columns


# ---------------------------------------------------------------------------
# Embedder dimension mismatch
# ---------------------------------------------------------------------------

async def test_index_raises_on_dimension_mismatch(store, table_with_data):
    """Embedder.dimension=4 but fn returns 2-dim vectors → ValueError."""
    async def wrong_dim_fn(texts: list[str]) -> list[list[float]]:
        return [[0.0, 1.0] for _ in texts]

    wrong_embedder = Embedder(fn=wrong_dim_fn, dimension=4)

    engine = SearchEngine(db_url="sqlite+aiosqlite://")
    engine._store = store

    config = TableConfig(
        table="docs",
        embeddings=[EmbeddingConfig(key="body_emb", source_column="body")],
    )
    await engine._ensure_config_table()

    with patch("dot_search.engine.get_embedder", return_value=wrong_embedder):
        with pytest.raises(ValueError, match="dimension mismatch"):
            await engine._run_indexing(config)


# ---------------------------------------------------------------------------
# Column existence checks
# ---------------------------------------------------------------------------

async def _make_engine(store):
    engine = SearchEngine(db_url="sqlite+aiosqlite://")
    engine._store = store
    await engine._ensure_config_table()
    return engine


async def test_index_raises_on_nonexistent_source_column(store, table_with_data):
    engine = await _make_engine(store)
    config = TableConfig(
        table="docs",
        embeddings=[EmbeddingConfig(key="body", source_column="nonexistent")],
    )
    with pytest.raises(ValueError, match="source_column 'nonexistent'.*does not exist"):
        await engine._run_indexing(config)


async def test_index_raises_on_nonexistent_bm25_column(store, table_with_data):
    engine = await _make_engine(store)
    config = TableConfig(
        table="docs",
        embeddings=[EmbeddingConfig(key="body", source_column="body")],
        bm25=[BM25Config(key="missing", source_column="nonexistent")],
    )
    with pytest.raises(ValueError, match="BM25Config source_column 'nonexistent'.*does not exist"):
        await engine._run_indexing(config)


async def test_index_accepts_valid_columns(store, table_with_data):
    engine = await _make_engine(store)
    config = TableConfig(
        table="docs",
        embeddings=[EmbeddingConfig(key="body", source_column="body")],
    )
    with patch("dot_search.engine.get_embedder", return_value=FAKE_EMBEDDER):
        await engine._run_indexing(config)



"""
Unit tests for PgStore directly.
Requires PostgreSQL + pgvector + ParadeDB.
"""
import pytest
import sqlalchemy as sa
from conftest import DB_URL
from dot_search.store import PgStore, SubSearch


@pytest.fixture
async def store():
    s = PgStore(db_url=DB_URL)
    yield s
    await s.close()


@pytest.fixture
async def table_with_data(store):
    """Creates a bare table and populates it for store-level tests."""
    async with store._engine.begin() as conn:
        await conn.execute(sa.text(
            """
            CREATE TABLE IF NOT EXISTS store_test (
                id TEXT PRIMARY KEY,
                body TEXT
            )
            """
        ))
        await conn.execute(sa.text(
            """
            INSERT INTO store_test (id, body) VALUES
            ('doc1', 'machine à laver en panne'),
            ('doc2', 'réparation de la chaudière'),
            ('doc3', 'inspection annuelle du système hydraulique')
            ON CONFLICT DO NOTHING
            """
        ))
    yield
    async with store._engine.begin() as conn:
        await conn.execute(sa.text("DROP TABLE IF EXISTS store_test"))


async def test_add_embedding_column_creates_column(store, table_with_data):
    await store.add_embedding_column("store_test", "ds_body_emb", dimension=4)
    async with store._engine.connect() as conn:
        result = await conn.execute(sa.text(
            "SELECT column_name FROM information_schema.columns "
            "WHERE table_name = 'store_test' AND column_name = 'ds_body_emb'"
        ))
        assert result.fetchone() is not None


async def test_add_embedding_column_is_idempotent(store, table_with_data):
    await store.add_embedding_column("store_test", "ds_body_emb", dimension=4)
    await store.add_embedding_column("store_test", "ds_body_emb", dimension=4)  # no error


async def test_write_embeddings(store, table_with_data):
    await store.add_embedding_column("store_test", "ds_body_emb", dimension=4)
    rows = [
        ("doc1", [1.0, 0.0, 0.0, 0.0]),
        ("doc2", [0.0, 1.0, 0.0, 0.0]),
        ("doc3", [0.0, 0.0, 1.0, 0.0]),
    ]
    await store.write_embeddings("store_test", "ds_body_emb", "id", rows)

    async with store._engine.connect() as conn:
        result = await conn.execute(sa.text(
            "SELECT id FROM store_test WHERE ds_body_emb IS NOT NULL"
        ))
        ids = {r[0] for r in result.fetchall()}
    assert ids == {"doc1", "doc2", "doc3"}


async def test_search_returns_results(store, table_with_data):
    await store.add_embedding_column("store_test", "ds_body_emb", dimension=4)
    await store.write_embeddings(
        "store_test", "ds_body_emb", "id",
        [
            ("doc1", [1.0, 0.0, 0.0, 0.0]),
            ("doc2", [0.0, 1.0, 0.0, 0.0]),
            ("doc3", [0.0, 0.0, 1.0, 0.0]),
        ],
    )
    subsearches = [
        SubSearch(
            embedding_column="ds_body_emb",
            query_embedding=[1.0, 0.0, 0.0, 0.0],  # closest to doc1
            weight=1.0,
        )
    ]
    results_per_subsearch = await store.search(
        table="store_test",
        pk_column="id",
        subsearches=subsearches,
        bm25searches=[],
        exact_searches=[],
        where=None,
        limit=3,
    )
    assert len(results_per_subsearch) == 1
    results = results_per_subsearch[0]
    assert len(results) > 0
    assert results[0].id == "doc1"


async def test_search_with_where_filters(store, table_with_data):
    await store.add_embedding_column("store_test", "ds_body_emb", dimension=4)
    await store.write_embeddings(
        "store_test", "ds_body_emb", "id",
        [
            ("doc1", [1.0, 0.0, 0.0, 0.0]),
            ("doc2", [1.0, 0.0, 0.0, 0.0]),
            ("doc3", [1.0, 0.0, 0.0, 0.0]),
        ],
    )
    subsearches = [
        SubSearch(
            embedding_column="ds_body_emb",
            query_embedding=[1.0, 0.0, 0.0, 0.0],
            weight=1.0,
        )
    ]
    results_per_subsearch = await store.search(
        table="store_test",
        pk_column="id",
        subsearches=subsearches,
        bm25searches=[],
        exact_searches=[],
        where="id = 'doc1'",
        limit=10,
    )
    results = results_per_subsearch[0]
    assert all(r.id == "doc1" for r in results)



from dot_search.engine import SearchEngine
from dot_search.types import (
    BM25Config,
    ExactConfig,
    EmbeddingConfig,
    SearchConfig,
    TableConfig,
    SearchResult,
)

__all__ = [
    "SearchEngine",
    "BM25Config",
    "ExactConfig",
    "EmbeddingConfig",
    "SearchConfig",
    "TableConfig",
    "SearchResult",
]



from __future__ import annotations

from typing import Any

import sqlalchemy as sa
from sqlalchemy.ext.asyncio import create_async_engine

from dot_search.store.protocol import BM25Search, ExactSearch, SubSearch, validate_embedding_column
from dot_search.types import SearchResult


class PgStore:
    def __init__(self, db_url: str) -> None:
        self._engine = create_async_engine(db_url)
        self._payload_cols_cache: dict[str, list[str]] = {}

    async def _get_payload_cols(self, conn: Any, table: str) -> list[str]:
        if table not in self._payload_cols_cache:
            result = await conn.execute(
                sa.text("SELECT column_name FROM information_schema.columns WHERE table_name = :table ORDER BY ordinal_position"),
                {"table": table},
            )
            self._payload_cols_cache[table] = [r[0] for r in result if not r[0].startswith("ds_")]
        return self._payload_cols_cache[table]

    async def add_embedding_column(self, table: str, column: str, dimension: int) -> None:
        validate_embedding_column(column)
        async with self._engine.begin() as conn:
            await conn.execute(sa.text("CREATE EXTENSION IF NOT EXISTS vector"))
            # Add column if not exists
            await conn.execute(sa.text(
                f"ALTER TABLE {table} ADD COLUMN IF NOT EXISTS {column} vector({dimension})"
            ))

    async def add_text_column(self, table: str, column: str) -> None:
        async with self._engine.begin() as conn:
            await conn.execute(sa.text(
                f"ALTER TABLE {table} ADD COLUMN IF NOT EXISTS {column} TEXT"
            ))

    async def write_texts(self, table: str, column: str, pk_column: str, rows: list[tuple[Any, str]]) -> None:
        if not rows:
            return
        async with self._engine.begin() as conn:
            await conn.execute(
                sa.text(f"UPDATE {table} SET {column} = :text WHERE {pk_column} = :id"),
                [{"text": text, "id": id_} for id_, text in rows],
            )

    async def add_bm25_index(self, table: str, columns: list[str]) -> None:
        index_name = f"{table}_bm25_idx"
        async with self._engine.begin() as conn:
            pk_result = await conn.execute(sa.text(
                f"""
                SELECT a.attname
                FROM pg_index i
                JOIN pg_attribute a ON a.attrelid = i.indrelid AND a.attnum = ANY(i.indkey)
                WHERE i.indrelid = '{table}'::regclass AND i.indisprimary
                LIMIT 1
                """
            ))
            pk_row = pk_result.fetchone()
            id_col = pk_row[0] if pk_row else "id"

            all_cols = ", ".join([id_col] + columns)
            await conn.execute(sa.text(f"DROP INDEX IF EXISTS {index_name}"))
            await conn.execute(sa.text(
                f"""
                CREATE INDEX {index_name} ON {table}
                USING bm25 ({all_cols})
                WITH (key_field='{id_col}')
                """
            ))

    async def write_embeddings(
        self,
        table: str,
        column: str,
        pk_column: str,
        rows: list[tuple[Any, list[float]]],
    ) -> None:
        if not rows:
            return
        async with self._engine.begin() as conn:
            await conn.execute(
                sa.text(f"UPDATE {table} SET {column} = :vec WHERE {pk_column} = :id"),
                [{"vec": str(vec), "id": id_} for id_, vec in rows],
            )

    async def read_column(
        self,
        table: str,
        pk_column: str,
        source_column: str,
    ) -> list[tuple[Any, str]]:
        async with self._engine.connect() as conn:
            result = await conn.execute(sa.text(
                f"SELECT {pk_column}, {source_column} FROM {table}"
            ))
            return [(row[0], str(row[1]) if row[1] is not None else "") for row in result]

    async def search(
        self,
        table: str,
        pk_column: str,
        subsearches: list[SubSearch],
        bm25searches: list[BM25Search],
        exact_searches: list[ExactSearch],
        where: str | None,
        limit: int,
    ) -> list[list[SearchResult]]:
        where_clause = f"AND {where}" if where else ""
        results = []

        async with self._engine.connect() as conn:
            payload_cols = await self._get_payload_cols(conn, table)
            select_cols = ", ".join(payload_cols)

            for sub in subsearches:
                vec_str = str(sub.query_embedding)
                sql = sa.text(
                    f"""
                    SELECT {select_cols},
                           1 - ({sub.embedding_column} <=> '{vec_str}'::vector) AS score
                    FROM {table}
                    WHERE {sub.embedding_column} IS NOT NULL
                    {where_clause}
                    ORDER BY {sub.embedding_column} <=> '{vec_str}'::vector
                    LIMIT :limit
                    """
                )
                result = await conn.execute(sql, {"limit": limit})
                rows = result.fetchall()
                col_names = list(result.keys())

                search_results = []
                for row in rows:
                    row_dict = dict(zip(col_names, row))
                    score = row_dict.pop("score")
                    id_val = row_dict[pk_column]
                    search_results.append(SearchResult(id=id_val, score=float(score), row=row_dict))
                results.append(search_results)

            for bm25 in bm25searches:
                bm25_query = f'{bm25.column}:"{bm25.query}"'
                sql = sa.text(
                    f"""
                    SELECT {select_cols},
                           paradedb.score({pk_column}) AS score
                    FROM {table}
                    WHERE {table} @@@ '{bm25_query}'
                    {where_clause}
                    ORDER BY score DESC
                    LIMIT :limit
                    """
                )
                result = await conn.execute(sql, {"limit": limit})
                rows = result.fetchall()
                col_names = list(result.keys())

                search_results = []
                for row in rows:
                    row_dict = dict(zip(col_names, row))
                    score = row_dict.pop("score")
                    id_val = row_dict[pk_column]
                    search_results.append(SearchResult(id=id_val, score=float(score), row=row_dict))
                results.append(search_results)

            for exact in exact_searches:
                sql = sa.text(
                    f"""
                    SELECT {select_cols}
                    FROM {table}
                    WHERE {exact.column} ILIKE :pattern
                    {where_clause}
                    LIMIT :limit
                    """
                )
                result = await conn.execute(sql, {"pattern": f"%{exact.query}%", "limit": limit})
                rows = result.fetchall()
                col_names = list(result.keys())

                search_results = []
                for row in rows:
                    row_dict = dict(zip(col_names, row))
                    id_val = row_dict[pk_column]
                    search_results.append(SearchResult(id=id_val, score=1.0, row=row_dict))
                results.append(search_results)

        return results

    async def get_columns(self, table: str) -> list[str]:
        async with self._engine.connect() as conn:
            result = await conn.execute(
                sa.text("SELECT column_name FROM information_schema.columns WHERE table_name = :table ORDER BY ordinal_position"),
                {"table": table},
            )
            return [r[0] for r in result]

    async def ensure_config_table(self) -> None:
        async with self._engine.begin() as conn:
            await conn.execute(sa.text(
                """
                CREATE TABLE IF NOT EXISTS ds_configs (
                    table_name TEXT PRIMARY KEY,
                    config_json TEXT NOT NULL
                )
                """
            ))

    async def save_config(self, index_id: str, config_json: str) -> None:
        async with self._engine.begin() as conn:
            await conn.execute(sa.text(
                """
                INSERT INTO ds_configs (table_name, config_json)
                VALUES (:index_id, :config_json)
                ON CONFLICT (table_name) DO UPDATE SET config_json = EXCLUDED.config_json
                """
            ), {"index_id": index_id, "config_json": config_json})

    async def load_config(self, index_id: str) -> str | None:
        async with self._engine.connect() as conn:
            result = await conn.execute(
                sa.text("SELECT config_json FROM ds_configs WHERE table_name = :index_id"),
                {"index_id": index_id},
            )
            row = result.fetchone()
            return row[0] if row else None

    async def close(self) -> None:
        await self._engine.dispose()



from dot_search import SearchResult
from dot_search.fusion import reciprocal_rank_fusion


def make_results(ids: list[str]) -> list[SearchResult]:
    return [SearchResult(id=id_, score=1.0, row={}) for id_ in ids]


def test_single_list_preserves_order():
    results = make_results(["a", "b", "c"])
    fused = reciprocal_rank_fusion([results], weights=[1.0])
    assert [r.id for r in fused] == ["a", "b", "c"]


def test_two_identical_lists_preserves_order():
    results = make_results(["a", "b", "c"])
    fused = reciprocal_rank_fusion([results, results], weights=[1.0, 1.0])
    assert [r.id for r in fused] == ["a", "b", "c"]


def test_two_lists_agreement_boosts_score():
    list1 = make_results(["a", "b", "c"])
    list2 = make_results(["b", "a", "c"])
    fused = reciprocal_rank_fusion([list1, list2], weights=[1.0, 1.0])
    ids = [r.id for r in fused]
    # "a" is rank 0 in list1, rank 1 in list2
    # "b" is rank 1 in list1, rank 0 in list2
    # Both should score higher than "c" which is rank 2 in both
    assert ids.index("c") > ids.index("a")
    assert ids.index("c") > ids.index("b")


def test_higher_weight_boosts_list():
    list1 = make_results(["a", "b"])  # "a" is top in list1
    list2 = make_results(["b", "a"])  # "b" is top in list2
    # Give list2 much higher weight — "b" should win
    fused = reciprocal_rank_fusion([list1, list2], weights=[1.0, 10.0])
    assert fused[0].id == "b"


def test_item_only_in_one_list_still_appears():
    list1 = make_results(["a", "b"])
    list2 = make_results(["c", "d"])
    fused = reciprocal_rank_fusion([list1, list2], weights=[1.0, 1.0])
    ids = [r.id for r in fused]
    assert set(ids) == {"a", "b", "c", "d"}


def test_empty_lists():
    fused = reciprocal_rank_fusion([[], []], weights=[1.0, 1.0])
    assert fused == []


def test_scores_are_positive():
    results = make_results(["a", "b", "c"])
    fused = reciprocal_rank_fusion([results], weights=[1.0])
    assert all(r.score > 0 for r in fused)


def test_scores_are_descending():
    results = make_results(["a", "b", "c", "d"])
    fused = reciprocal_rank_fusion([results], weights=[1.0])
    scores = [r.score for r in fused]
    assert scores == sorted(scores, reverse=True)



"""
SQLiteStore tests — in-memory SQLite, no Postgres required.
"""
import pytest
from dot_search.store.sqlite import SQLiteStore
from dot_search.store.protocol import ExactSearch, SubSearch


SQLITE_URL = "sqlite+aiosqlite://"


@pytest.fixture
async def store():
    s = SQLiteStore(db_url=SQLITE_URL)
    yield s
    await s.close()


@pytest.fixture
async def table_with_data(store):
    conn = await store._get_conn()
    await conn.execute("CREATE TABLE IF NOT EXISTS docs (id TEXT PRIMARY KEY, body TEXT)")
    await conn.execute("INSERT OR IGNORE INTO docs (id, body) VALUES ('d1', 'hello world'), ('d2', 'foo bar'), ('d3', 'baz qux')")
    await conn.commit()
    yield
    await conn.execute("DROP TABLE IF EXISTS docs")
    await conn.commit()


async def test_add_embedding_column(store, table_with_data):
    await store.add_embedding_column("docs", "ds_body_emb", dimension=4)
    conn = await store._get_conn()
    cursor = await conn.execute("PRAGMA table_info(docs)")
    rows = await cursor.fetchall()
    columns = {row[1] for row in rows}
    assert "ds_body_emb" in columns


async def test_write_and_read_embeddings(store, table_with_data):
    await store.add_embedding_column("docs", "ds_body_emb", dimension=4)
    rows = [
        ("d1", [1.0, 0.0, 0.0, 0.0]),
        ("d2", [0.0, 1.0, 0.0, 0.0]),
        ("d3", [0.0, 0.0, 1.0, 0.0]),
    ]
    await store.write_embeddings("docs", "ds_body_emb", "id", rows)

    conn = await store._get_conn()
    cursor = await conn.execute(
        "SELECT id FROM docs WHERE ds_body_emb IS NOT NULL"
    )
    fetched = await cursor.fetchall()
    ids = {r[0] for r in fetched}
    assert ids == {"d1", "d2", "d3"}


async def test_read_column_with_source(store, table_with_data):
    rows = await store.read_column("docs", "id", "body")
    assert len(rows) == 3
    assert all(isinstance(text, str) for _, text in rows)



async def test_search_returns_closest(store, table_with_data):
    await store.add_embedding_column("docs", "ds_body_emb", dimension=4)
    await store.write_embeddings(
        "docs", "ds_body_emb", "id",
        [
            ("d1", [1.0, 0.0, 0.0, 0.0]),
            ("d2", [0.0, 1.0, 0.0, 0.0]),
            ("d3", [0.0, 0.0, 1.0, 0.0]),
        ],
    )
    subsearches = [
        SubSearch(
            embedding_column="ds_body_emb",
            query_embedding=[1.0, 0.0, 0.0, 0.0],
            weight=1.0,
        )
    ]
    results_per_sub = await store.search(
        table="docs",
        pk_column="id",
        subsearches=subsearches,
        bm25searches=[],
        exact_searches=[],
        where=None,
        limit=3,
    )
    assert len(results_per_sub) == 1
    results = results_per_sub[0]
    assert results[0].id == "d1"


async def test_add_bm25_index_is_noop(store, table_with_data):
    # Should not raise
    await store.add_bm25_index("docs", ["body"])


async def test_exact_search_matches_substring(store, table_with_data):
    results_per_search = await store.search(
        table="docs",
        pk_column="id",
        subsearches=[],
        bm25searches=[],
        exact_searches=[ExactSearch(column="body", query="hello")],
        where=None,
        limit=10,
    )
    assert len(results_per_search) == 1
    results = results_per_search[0]
    assert len(results) == 1
    assert results[0].id == "d1"


async def test_exact_search_no_match_returns_empty(store, table_with_data):
    results_per_search = await store.search(
        table="docs",
        pk_column="id",
        subsearches=[],
        bm25searches=[],
        exact_searches=[ExactSearch(column="body", query="zzznomatch")],
        where=None,
        limit=10,
    )
    assert results_per_search[0] == []


async def test_exact_search_respects_where(store, table_with_data):
    # "hello world" matches d1, but filter excludes it
    results_per_search = await store.search(
        table="docs",
        pk_column="id",
        subsearches=[],
        bm25searches=[],
        exact_searches=[ExactSearch(column="body", query="hello")],
        where="id != 'd1'",
        limit=10,
    )
    assert results_per_search[0] == []


async def test_exact_search_multiple_columns_produce_separate_lists(store, table_with_data):
    results_per_search = await store.search(
        table="docs",
        pk_column="id",
        subsearches=[],
        bm25searches=[],
        exact_searches=[
            ExactSearch(column="body", query="hello"),
            ExactSearch(column="body", query="foo"),
        ],
        where=None,
        limit=10,
    )
    assert len(results_per_search) == 2
    assert results_per_search[0][0].id == "d1"
    assert results_per_search[1][0].id == "d2"


async def test_exact_search_scores_are_all_one(store, table_with_data):
    results_per_search = await store.search(
        table="docs",
        pk_column="id",
        subsearches=[],
        bm25searches=[],
        exact_searches=[ExactSearch(column="body", query="o")],  # matches d1 (hello world), d2 (foo bar)
        where=None,
        limit=10,
    )
    results = results_per_search[0]
    assert all(r.score == 1.0 for r in results)



from dot_search.store.pg import PgStore
from dot_search.store.protocol import SubSearch
from dot_search.store.sqlite import SQLiteStore


def make_store(db_url: str) -> PgStore | SQLiteStore:
    if db_url.startswith("postgresql"):
        return PgStore(db_url)
    elif db_url.startswith("sqlite"):
        return SQLiteStore(db_url)
    raise ValueError(f"Unsupported db_url scheme: {db_url}")



import asyncio
from dot_search import SearchEngine, TableConfig, EmbeddingConfig, BM25Config, ExactConfig


async def main():
    engine = SearchEngine(db_url="postgresql+asyncpg://user:pass@localhost/mydb")

    # --- Index: multiple embeddings + BM25 + exact ---
    await engine.index(config=TableConfig(
        table="articles",
        pk_column="id",                 # defaults to "id" if omitted
        batch_size=100,                 # defaults to 500 if omitted
        embeddings=[
            EmbeddingConfig(
                source_column="body",
                target_column="ds_body",
            ),
            EmbeddingConfig(
                source_column="title",
                target_column="ds_title",
            ),
        ],
        bm25=[
            BM25Config(source_column="title"),
            BM25Config(source_column="body"),
        ],
        exact=[ExactConfig(source_column="name")],
    ))

    # --- Index 2: title-only, separate index_id on the same table ---
    await engine.index(config=TableConfig(
        table="articles",
        index_id="article_titles",
        embeddings=[
            EmbeddingConfig(
                source_column="title",
                target_column="ds_article_titles_title",
            ),
        ],
    ))

    # --- Re-index when new rows arrive ---
    await engine.index(index_id="articles")        # reloads saved config, re-embeds all rows
    await engine.index(index_id="article_titles")

    # --- Search: hybrid (vector + BM25 + exact), SQL filter, per-query weight overrides ---
    print("=== Hybrid search ===")
    results = await engine.search(
        "fermentation and gut health",
        "articles",
        limit=10,
        where="published_year >= 2022 AND topic = 'health'",
        embedding_weights={
            "ds_body": 1.0,
            "ds_title": 0.3,
        },
        bm25_weights={"title": 0.5, "body": 2.0},
    )
    for r in results:
        print(r)

    # --- Search: BM25-only ---
    print("=== BM25-only search ===")
    results = await engine.search("fermentation", "articles", strategy="bm25", limit=5)
    for r in results:
        print(r)

    # --- Search: exact-only ---
    print("=== Exact search ===")
    results = await engine.search("Dupont", "articles", strategy="exact", limit=5)
    for r in results:
        print(r)

    # --- Search: title-only index ---
    print("=== Title-only index search ===")
    results = await engine.search("gut health", "article_titles", limit=5)
    for r in results:
        print(r)


asyncio.run(main())



"""
Integration tests for SearchEngine.
Requires a running PostgreSQL instance with pgvector and ParadeDB extensions.
Connection: DB_URL from conftest.py
"""
from unittest.mock import patch

import pytest
import sqlalchemy as sa
from conftest import DB_URL, fake_embed_fn
from dot_search import BM25Config, EmbeddingConfig, SearchConfig, TableConfig, SearchEngine
from dot_search.types import ExactConfig
from dot_search.embed import Embedder


FAKE_EMBEDDER = Embedder(fn=fake_embed_fn, dimension=4)


@pytest.fixture
async def engine():
    e = SearchEngine(db_url=DB_URL)
    yield e
    await e.close()


@pytest.fixture
async def patient_table(engine):
    """Creates a fake 'patient' table mimicking a Django-managed table."""
    async with engine._store._engine.begin() as conn:
        await conn.execute(sa.text(
            """
            CREATE TABLE IF NOT EXISTS patient (
                id SERIAL PRIMARY KEY,
                name TEXT NOT NULL,
                age INTEGER,
                status TEXT,
                comment TEXT
            )
            """
        ))
        await conn.execute(sa.text(
            """
            INSERT INTO patient (name, age, status, comment) VALUES
            ('Dupont', 34, 'active', 'Patient calme, coopératif, suit bien les consignes.'),
            ('Martin', 52, 'active', 'Difficile à gérer, refuse souvent les soins.'),
            ('Bernard', 28, 'inactive', 'Très motivé, bonne progression thérapeutique.'),
            ('Leroy', 61, 'active', 'Patient anxieux mais de bonne volonté.'),
            ('Moreau', 45, 'active', 'Comportement imprévisible, nécessite un suivi renforcé.')
            ON CONFLICT DO NOTHING
            """
        ))
    yield
    async with engine._store._engine.begin() as conn:
        await conn.execute(sa.text("DROP TABLE IF EXISTS patient"))
        await conn.execute(sa.text("DROP TABLE IF EXISTS ds_configs"))


async def test_augment_adds_ds_column(engine, patient_table):
    config = TableConfig(
        table="patient",
        embeddings=[
            EmbeddingConfig(
                key="comment_default",
                source_column="comment",
            )
        ],
    )
    with patch("dot_search.engine.get_embedder", return_value=FAKE_EMBEDDER):
        await engine.index(config)
    async with engine._store._engine.connect() as conn:
        result = await conn.execute(sa.text(
            "SELECT ds_comment_default FROM patient WHERE ds_comment_default IS NOT NULL"
        ))
        rows = result.fetchall()
    assert len(rows) == 5


async def test_augment_does_not_alter_existing_columns(engine, patient_table):
    config = TableConfig(
        table="patient",
        embeddings=[
            EmbeddingConfig(
                key="comment_default",
                source_column="comment",
            )
        ],
    )
    with patch("dot_search.engine.get_embedder", return_value=FAKE_EMBEDDER):
        await engine.index(config)
    async with engine._store._engine.connect() as conn:
        result = await conn.execute(sa.text(
            "SELECT id, name, age, status, comment FROM patient ORDER BY id"
        ))
        rows = result.fetchall()
    assert len(rows) == 5
    assert rows[0].name == "Dupont"


async def test_search_with_where_filter(engine, patient_table):
    config = TableConfig(
        table="patient",
        embeddings=[
            EmbeddingConfig(
                key="comment_default",
                source_column="comment",
            )
        ],
    )
    with patch("dot_search.engine.get_embedder", return_value=FAKE_EMBEDDER):
        await engine.index(config)
        results = await engine.search(
            "patient facile à gérer",
            "patient",
            SearchConfig(where="status = 'active' AND age < 50"),
        )
    returned_names = {r.row["name"] for r in results}
    assert "Bernard" not in returned_names   # inactive
    assert "Martin" not in returned_names    # age 52


async def test_search_results_contain_row_data(engine, patient_table):
    config = TableConfig(
        table="patient",
        embeddings=[
            EmbeddingConfig(
                key="comment_default",
                source_column="comment",
            )
        ],
    )
    with patch("dot_search.engine.get_embedder", return_value=FAKE_EMBEDDER):
        await engine.index(config)
        results = await engine.search(
            "patient",
            "patient",
            SearchConfig(limit=3),
        )
    for r in results:
        assert "name" in r.row
        assert "comment" in r.row
        assert "ds_comment_default" not in r.row


async def test_multi_embedding_columns(engine, patient_table):
    """Two embedding targets → two ds_ columns created independently."""
    config = TableConfig(
        table="patient",
        embeddings=[
            EmbeddingConfig(
                key="comment_multilingual",
                source_column="comment",
            ),
            EmbeddingConfig(
                key="comment_longctx",
                source_column="comment",
            ),
        ],
    )
    with patch("dot_search.engine.get_embedder", return_value=FAKE_EMBEDDER):
        await engine.index(config)
    async with engine._store._engine.connect() as conn:
        result = await conn.execute(sa.text(
            "SELECT ds_comment_multilingual, ds_comment_longctx FROM patient"
        ))
        rows = result.fetchall()
    assert all(r.ds_comment_multilingual is not None for r in rows)
    assert all(r.ds_comment_longctx is not None for r in rows)


async def test_multiple_bm25_config_round_trip(engine, patient_table):
    """Config with two BM25Configs serializes and deserializes correctly."""
    config = TableConfig(
        table="patient",
        embeddings=[
            EmbeddingConfig(
                key="comment_default",
                source_column="comment",
            )
        ],
        bm25=[
            BM25Config(key="name", source_column="name"),
            BM25Config(key="comment", source_column="comment"),
        ],
    )
    await engine._ensure_config_table()
    await engine._save_config(config)

    loaded = await engine._load_config("patient")
    assert len(loaded.bm25) == 2
    assert loaded.bm25[0].source_column == "name"
    assert loaded.bm25[1].source_column == "comment"


async def test_exact_search_returns_matching_rows(engine, patient_table):
    config = TableConfig(
        table="patient",
        embeddings=[],
        exact=[ExactConfig(key="name", source_column="name")],
    )
    with patch("dot_search.engine.get_embedder", return_value=FAKE_EMBEDDER):
        await engine.index(config)
        results = await engine.search("Dupont", "patient", SearchConfig(strategy="exact"))
    assert len(results) == 1
    assert results[0].row["name"] == "Dupont"


async def test_exact_search_infix_match(engine, patient_table):
    config = TableConfig(
        table="patient",
        embeddings=[],
        exact=[ExactConfig(key="comment", source_column="comment")],
    )
    with patch("dot_search.engine.get_embedder", return_value=FAKE_EMBEDDER):
        await engine.index(config)
        results = await engine.search("suivi", "patient", SearchConfig(strategy="exact"))
    assert len(results) == 1
    assert results[0].row["name"] == "Moreau"


async def test_exact_search_no_match_returns_empty(engine, patient_table):
    config = TableConfig(
        table="patient",
        embeddings=[],
        exact=[ExactConfig(key="name", source_column="name")],
    )
    with patch("dot_search.engine.get_embedder", return_value=FAKE_EMBEDDER):
        await engine.index(config)
        results = await engine.search("zzznomatch", "patient", SearchConfig(strategy="exact"))
    assert results == []


async def test_exact_search_raises_without_exact_config(engine, patient_table):
    config = TableConfig(
        table="patient",
        embeddings=[EmbeddingConfig(key="comment_default", source_column="comment")],
    )
    with patch("dot_search.engine.get_embedder", return_value=FAKE_EMBEDDER):
        await engine.index(config)
        with pytest.raises(ValueError, match="no exact config"):
            await engine.search("Dupont", "patient", SearchConfig(strategy="exact"))


async def test_exact_config_round_trip(engine, patient_table):
    config = TableConfig(
        table="patient",
        embeddings=[],
        exact=[ExactConfig(key="name", source_column="name")],
    )
    await engine._ensure_config_table()
    await engine._save_config(config)
    loaded = await engine._load_config("patient")
    assert len(loaded.exact) == 1
    assert loaded.exact[0].source_column == "name"



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Protocol

from dot_search.types import SearchResult


@dataclass
class SubSearch:
    embedding_column: str
    query_embedding: list[float]
    weight: float = 1.0


@dataclass
class BM25Search:
    column: str
    query: str
    weight: float = 1.0


@dataclass
class ExactSearch:
    column: str
    query: str
    weight: float = 1.0


def validate_embedding_column(column: str) -> None:
    if not column.startswith("ds_"):
        raise ValueError(f"Embedding column must start with 'ds_', got: {column!r}")


class Store(Protocol):
    async def add_embedding_column(self, table: str, column: str, dimension: int) -> None: ...
    async def add_text_column(self, table: str, column: str) -> None: ...
    async def write_texts(self, table: str, column: str, pk_column: str, rows: list[tuple[Any, str]]) -> None: ...
    async def add_bm25_index(self, table: str, columns: list[str]) -> None: ...
    async def write_embeddings(
        self,
        table: str,
        column: str,
        pk_column: str,
        rows: list[tuple[Any, list[float]]],
    ) -> None: ...
    async def read_column(
        self,
        table: str,
        pk_column: str,
        source_column: str,
    ) -> list[tuple[Any, str]]: ...
    async def search(
        self,
        table: str,
        pk_column: str,
        subsearches: list[SubSearch],
        bm25searches: list[BM25Search],
        exact_searches: list[ExactSearch],
        where: str | None,
        limit: int,
    ) -> list[list[SearchResult]]: ...
    async def get_columns(self, table: str) -> list[str]: ...
    async def ensure_config_table(self) -> None: ...
    async def save_config(self, index_id: str, config_json: str) -> None: ...
    async def load_config(self, index_id: str) -> str | None: ...
    async def close(self) -> None: ...



from __future__ import annotations

import json
from dataclasses import dataclass, field
from typing import Any, Literal

SearchStrategy = Literal["bm25", "vector", "hybrid", "exact"]


@dataclass
class EmbeddingConfig:
    key: str
    source_column: str
    provider: str = "default"
    model: str = "default"

    @property
    def target_column(self) -> str:
        return f"ds_{self.key}"


@dataclass
class BM25Config:
    key: str
    source_column: str


@dataclass
class ExactConfig:
    key: str
    source_column: str


@dataclass
class TableConfig:
    table: str
    embeddings: list[EmbeddingConfig]
    pk_column: str = "id"
    bm25: list[BM25Config] = field(default_factory=list)
    exact: list[ExactConfig] = field(default_factory=list)
    batch_size: int = 500
    index_id: str = ""

    def __post_init__(self) -> None:
        if not self.index_id:
            self.index_id = self.table
        all_keys = (
            [e.key for e in self.embeddings]
            + [b.key for b in self.bm25]
            + [e.key for e in self.exact]
        )
        dupes = [k for k in all_keys if all_keys.count(k) > 1]
        if dupes:
            raise ValueError(f"Duplicate keys in TableConfig: {sorted(set(dupes))}")

    @classmethod
    def from_json(cls, raw: str) -> TableConfig:
        data = json.loads(raw)
        data["embeddings"] = [EmbeddingConfig(**e) for e in data["embeddings"]]
        data["bm25"] = [BM25Config(**b) for b in data["bm25"]]
        data["exact"] = [ExactConfig(**e) for e in data.get("exact", [])]
        return cls(**data)


@dataclass
class SearchConfig:
    strategy: SearchStrategy = "hybrid"
    limit: int = 10
    where: str | None = None
    weights: dict[str, float] = field(default_factory=dict)


@dataclass
class SearchResult:
    id: Any
    score: float
    row: dict[str, Any]  # all non-ds_ columns



import asyncio
from dot_search import SearchEngine, TableConfig, EmbeddingConfig


async def main():
    engine = SearchEngine(db_url="postgresql+asyncpg://user:pass@localhost/mydb")

    # Index — config saved to ds_configs
    await engine.index(config=TableConfig(
        table="documents",
        embeddings=[EmbeddingConfig(source_column="body", target_column="ds_body")],
    ))

    # Search
    results = await engine.search("neural networks", "documents", limit=5)
    for r in results:
        print(r)


asyncio.run(main())
